﻿Imports System.Math

Public Class JointCalc


    'アーム先端座標から、各関節のＲＣサーボに出力するデータを求める
    Shared Sub coo2servo3ax(ByVal coo() As Double, ByRef servo_data() As Integer, ByRef joint_angle() As Double)
        Dim i As Integer
        Dim x, y, z, x2, x3, z3, r(2), coo_work(2, 2), coo_work1(3, 2) As Double

        x = coo(0)
        y = coo(1)
        z = coo(2)

        '****************************************************
        '台座Ｚ軸周りの角度を求める（ＲＣサーボ０の角度）
        joint_angle(0) = Atan2(y, x) / PI * 180

        x2 = (x ^ 2 + y ^ 2) ^ 0.5

        '****************************************************
        'ひじの座標を求める
        'crcirに与える変数を準備
        r(0) = frmUtils.bone_length(0, 0)
        r(1) = frmUtils.bone_length(0, 1)

        coo_work(0, 0) = 0 'x サーボ１の軸の位置
        coo_work(0, 1) = 0 'y
        coo_work(1, 0) = x2 'x　アーム先端
        coo_work(1, 1) = z 'y

        'アームの根本、ひじ、先端の３点で作る平面の中で２円の交点を計算する
        i = crcir(coo_work, r)

        'Ｚ座標の大きい方が関節可動範囲なので、ひじ関節の位置として選ぶ
        If coo_work(0, 1) > coo_work(1, 1) Then
            i = 0
        Else
            i = 1
        End If
        'ひじの座標
        x3 = coo_work(i, 0)
        z3 = coo_work(i, 1)

        joint_angle(1) = Atan2(-x3, z3) / PI * 180

        '****************************************************
        'ひじの角度を求める
        coo_work1(0, 0) = 0
        coo_work1(0, 1) = 0
        coo_work1(1, 0) = x3
        coo_work1(1, 1) = z3
        coo_work1(2, 0) = x2
        coo_work1(2, 1) = z

        joint_angle(2) = 90 + vector_dir_f(coo_work1) / PI * 180

        For i = 0 To 2
            servo_data(i) = angle2servo(i, joint_angle(i))
        Next

    End Sub

    '/******************************************************************/
    '/**  隣接するベクトルの一方からみた他方の角度を計算する          **/
    '/**  coo : 計算したいベクトルの座標　　　　　　　　　　          **/
    '/**  ベクトルはcoo[0][x]->coo[1][x]->coo[2][x]の順に接続するので **/
    '/**  実際には coo[1][x] での角度が求められる。　　　　           **/
    '/******************************************************************/
    Shared Function vector_dir_f(ByRef coo(,) As Double)

        Dim coof(2, 2), q1, q3 As Double

        '/* make vector BC*/
        coof(0, 0) = coo(2, 0) - coo(1, 0)
        coof(0, 1) = coo(2, 1) - coo(1, 1)

        '/* make vector AB*/
        coof(1, 0) = coo(1, 0) - coo(0, 0)
        coof(1, 1) = coo(1, 1) - coo(0, 1)


        '/* cluculate Q1 */
        q1 = Atan2(coof(1, 1), coof(1, 0))

        '/*************************************/
        '/* adjst Q2 to Q3 */
        rotate_1D(coof, 0.0, 0.0, -q1, 1)

        '/*************************************/
        '/* cluculate Q3 */
        q3 = Atan2(coof(0, 1), coof(0, 0))

        Return (q3)
    End Function

    Shared Function angle2servo(ByVal side_id As Integer, ByVal joint_id As Integer, ByVal joint_angle As Double) As Integer

        Dim i, j, servo_data As Integer
        Dim a, b, q_refp(3) As Double

        '参照点の角度をラジアンに変換
        j = 0
        For i = 0 To 2
            j = i * 2
            q_refp(i) = frmUtils.servo_st_en(side_id, joint_id, j) / 180.0 * PI
        Next i

        '補間直線の傾きを求める
        joint_angle = joint_angle / 180.0 * PI
        If joint_angle > q_refp(1) Then
            a = (frmUtils.servo_st_en(side_id, joint_id, 5) - frmUtils.servo_st_en(side_id, joint_id, 3)) _
                / (q_refp(2) - q_refp(1))
        Else
            a = (frmUtils.servo_st_en(side_id, joint_id, 3) - frmUtils.servo_st_en(side_id, joint_id, 1)) _
                / (q_refp(1) - q_refp(0))
        End If

        '補間直線のＹ切片を求める
        b = frmUtils.servo_st_en(side_id, joint_id, 3) - a * q_refp(1)

        servo_data = Int((a * joint_angle + b))

        Return (servo_data)
    End Function

    'joint_id を０～２１で与えるバージョン
    Shared Function angle2servo(ByVal joint_id As Integer, ByVal joint_angle As Double) As Integer

        Dim id, side_id, servo_data As Integer

        '目標角度のrawデータを求める
        side_id = 0
        id = joint_id

        servo_data = angle2servo(side_id, id, joint_angle)

        Return (servo_data)

    End Function
    '/*******************************************************************/
    '/**  回転の１次変換で任意の点を中心にして任意の点を回転する       **/
    '/**  coo[0][0] : x1  coo[0][1] : y1                               **/
    '/**  coo[1][0] : x2  coo[1][1] : y2                               **/
    '/**  canter_x, y : 回転変換の中心　                               **/
    '/**  direction   : 回転角　　　　　　                             **/
    '/**  data_num    : 回転するデータ数                               **/
    '/**  　　cooに、２次元配列の先頭を指すポインタを渡す              **/
    '/*******************************************************************/
    Shared Sub rotate_1D(ByRef coo(,) As Double, ByVal center_x As Double, _
                ByVal center_y As Double, ByVal direction As Double, ByVal data_num As Integer)

        Dim i As Integer
        Dim x, y As Double

        '/* 受け取った座標群を回転の中心が原点に来るように平行移動する */
        For i = 0 To data_num - 1
            coo(i, 0) = coo(i, 0) - center_x
            coo(i, 1) = coo(i, 1) - center_y
        Next

        '/* 原点を中心として回転の一次変換をする */
        For i = 0 To data_num - 1
            x = coo(i, 0)
            y = coo(i, 1)
            coo(i, 0) = x * Cos(direction) - y * Sin(direction)
            coo(i, 1) = x * Sin(direction) + y * Cos(direction)
        Next

        '/* 受け取った座標群を回転の中心が元の位置に来るように平行移動し戻す */
        For i = 0 To data_num - 1
            coo(i, 0) = coo(i, 0) + center_x
            coo(i, 1) = coo(i, 1) + center_y
        Next

    End Sub

    '/************************************************************************/
    '/*   ２円の交点の座標を求める                                           */
    '/*   coo[2][2]...２円の中心座標, 結果の交点座標もここへ返される　       */
    '/*   r[2]     ...２円の半径                                             */
    '/*   元々２円が第１象現にあるときのみ正しい結果が出るサブルーチン　　   */
    '/*   だったので２円の直径分だけオフセットを与えて計算している。　       */
    '/*   crcir とはパラメータの与え方がＸＹ逆なだけ　　　　　　　　　       */
    '/************************************************************************/
    Shared Function crcir(ByRef coo(,) As Double, ByVal r() As Double) As Integer
        Dim ii, jj As Integer
        Dim x0, y0, x1, y1, x5, y5 As Double
        Dim a, aa, b, e, f, h, i, l, l2, m, o, p As Double
        Dim a1, d As Double
        Dim vr0(2), vx(2), vy(2), r0, r1, r2, r3, cosq1, sinq1 As Double
        Dim offset, offset1 As Double


        '２円のX座標が同じ時の例外処理
        If coo(0, 0) = coo(1, 0) Then
            '		printf("例外処理が発生しました。crcirで０の割り算が発生しました。\n")
            '		printf("２円のx座標が同じになっています。\n")
            x0 = coo(0, 0)
            y0 = coo(0, 1)
            x1 = coo(1, 0)
            y1 = coo(1, 1)

            d = coo(1, 1) - coo(0, 1)
            a1 = ((r(0) * r(0)) - (r(1) * r(1)) + (d * d)) / (2 * d)
            aa = (r(0) * r(0)) - (a1 * a1)
            If aa < 0.0 Then
                'printf("２円の交点が求まりません(crcir)\n")
                Return (-1)
            End If

            b = Sqrt(aa)

            coo(0, 0) = x0 - b
            coo(0, 1) = y0 + a1

            coo(1, 0) = x0 + b
            coo(1, 1) = y0 + a1

            Return (0)
        End If


        offset1 = 0.0
        For ii = 0 To 1
            For jj = 0 To 1
                If coo(ii, jj) < offset1 Then offset1 = coo(ii, jj)
            Next
        Next


        offset = (r(0) + r(1)) * 2 - offset1

        h = coo(0, 0) + offset
        i = coo(0, 1) + offset
        e = coo(1, 0) + offset
        f = coo(1, 1) + offset
        r1 = r(0)
        r2 = r(1)
        r3 = r1 + r2

        vr0(0) = e - h
        vr0(1) = f - i
        r0 = Sqrt(vr0(0) * vr0(0) + vr0(1) * vr0(1))

        '/*まず交点が作成可能かチェック、ダメならエラー*/
        If r0 > r3 Then
            Return (-1)
        End If

        cosq1 = (r0 * r0 + r1 * r1 - r2 * r2) / (2 * r0 * r1)
        sinq1 = Sqrt(1 - cosq1 * cosq1)

        vx(0) = vr0(0) / r0
        vx(1) = vr0(1) / r0
        vy(0) = -vx(1)
        vy(1) = vx(0)

        l = r1 * cosq1 * vx(0) + r1 * sinq1 * vy(0) + h
        m = r1 * cosq1 * vx(1) + r1 * sinq1 * vy(1) + i

        coo(0, 0) = l - offset
        coo(0, 1) = m - offset

        a = vx(1) / vx(0)
        b = vy(1) / vy(0)
        x5 = (a * h - b * l + m - i) / (a - b)
        y5 = a * (x5 - h) + i
        a = x5 - l
        b = y5 - m
        l2 = Sqrt(a * a + b * b)
        o = -vy(0) * 2 * l2 + l
        p = -vy(1) * 2 * l2 + m
        '/* printf("x = %f   y = %f \n",o,p)
        coo(1, 0) = o - offset
        coo(1, 1) = p - offset
        Return (0)
    End Function
End Class
